package ga.base;

public class Population {
	Chromosome[] chrom;
	Evaluator evaluator;

	double mutation_rate;
	Mutation mutation;
	
	double crossover_rate;
	Crossover crossover;
	
	int generation;
	
	public Population(int populationSize, int chromosomeLength, Evaluator eval) {
		chrom = new Chromosome[populationSize];
		for( int i = 0 ; i < populationSize ; i++ ) {
			chrom[i] = new Chromosome(chromosomeLength);
		}
		evaluator = eval;
		mutation_rate = 0.0;
		mutation = null;
		crossover_rate = 0.0;
		crossover = null;
		generation = 0;
	}
	
	public void randomize() {
		for( Chromosome c : chrom ) {
			c.randomize();
		}
		evaluate();
	}
	
	public void update() {
		reproduce();
		doCrossoverOperation();
		doMutationOperation();
		evaluate();
		generation++;
	}
	
	public void setMutationOperation(Mutation m, double mrate) {
		mutation_rate = mrate;
		mutation = m;
	}
	
	public void setCrossoverOperation(Crossover c, double crate) {
		crossover_rate = crate;
		crossover = c;
	}
	
	public int getGeneration() {
		return generation;
	}
	
	public int getPopulationSize() {
		return chrom.length;
	}
	
	public Chromosome getChromosomeAtIndex(int i) {
		return chrom[i];
	}
	
	public Chromosome getChromosomeWithMaxFitnessValue() {
		return chrom[0];
	}
	
	public double getMaxFitnessValue() {
		return chrom[0].fitnessValue;
	}
	
	public double getAverageFitnessValue() {
		double sum = 0.0;
		for( int i = 0 ; i < chrom.length ; i++ ) {
			sum += chrom[0].getFitnessValue();
		}
		return sum / chrom.length;
	}
	
	private void doMutationOperation() {
		if( mutation != null ) {
			for( int i = 1 ; i < chrom.length ; i++ ) {
				mutation.mutate(chrom[i], mutation_rate);
			}
		}
	}
	
	private void doCrossoverOperation() {
		if( crossover != null ) {
			int[] idx = new int[chrom.length - 1];
			for( int i = 0 ; i < idx.length ; i++ ) {
				idx[i] = i + 1;
			}

			int num = (int)(chrom.length * crossover_rate);
			int idxlen = idx.length;
			int i = 0;
			while( i < num ) {
				int a = (int)(Math.random() *  idxlen     );
				int b = (int)(Math.random() * (idxlen - 1));
				if( a == b ) {
					b = idxlen - 1;
				}
				crossover.crossover(chrom[idx[a]], chrom[idx[b]]);
				idx[a] = idx[idxlen - 1];
				idx[b] = idx[idxlen - 2];
				idxlen -= 2;
				i += 2;
			}
		}
	}
	
	private void bubbleSort(Chromosome[] carr) {
		for( int i = 0 ; i < carr.length - 1 ; i++ ) {
			for( int j = i ; j < carr.length - 1 ; j++ ) {
				if( carr[j].getFitnessValue() < carr[j+1].getFitnessValue() ) {
					Chromosome tmp = carr[j];
					carr[j] = carr[j+1];
					carr[j+1] = tmp;
				}
			}
		}
	}
	
	private void evaluate() {
		for( Chromosome c : chrom ) {
			c.setFitnessValue(evaluator.evaluate(c));
		}
		bubbleSort(chrom);
	}
	
	private void reproduce() {
		if (chrom[chrom.length - 1].getFitnessValue() < chrom[0].getFitnessValue()) {
			double sum = 0.0;
			for (Chromosome c : chrom) {
				sum += c.getFitnessValue();
			}

			Chromosome[] newChrom = new Chromosome[chrom.length];
			newChrom[0] = new Chromosome(chrom[0]); // keep an elitest chromosome
			for (int i = 1; i < newChrom.length; i++) {
				newChrom[i] = new Chromosome(selectAChromosome(sum));
			}
			chrom = newChrom;
		}
	}
	
	private Chromosome selectAChromosome(double sum) {
		Chromosome ret = null;
		double r = Math.random() * sum;
		for( Chromosome c : chrom ) {
			if( r < c.getFitnessValue() ) {
				ret = c;
				break;
			} else {
				r -= c.getFitnessValue();
			}
		}
		return ret;
	}
}
